/*
 *
 * Copyright (c) 1988, 1999 Hellmuth Michaelis. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *---------------------------------------------------------------------------
 *
 *	history.c - commandline history and edit handling
 *      -------------------------------------------------
 *
 *	Last Edit-Date: [Fri Jul 30 11:04:06 1999]
 *
 *	$Id$
 *
 *----------------------------------------------------------------------------*/

#include "cwish.h"				/* local includes */

static char kbuf[HISLNLEN];			/* kill buffer */
static int kbuf_init = 0;			/* static kill buffer */
static char *beg;				/* line start      (first char on line) */
static char *pnt;				/* next insertion position     (cursor) */
static char *end;				/* line end    (first free pos on line) */
static char *sbeg;				/* first character on screen            */
static int lmax;				/* maximum charcount requested from caller */

static void pmstr(char *);			/* forward decl */
static void pnstr(int, char *);
static char *alias(char *command);

/*---------------------------------------------------------------------------*
 *	return current (cursor) column
 *---------------------------------------------------------------------------*/
int
curcol(void)
{
	return ((int) (pnt - sbeg));
}

/*---------------------------------------------------------------------------*
 *	return true, if commandline empty ( == operate on file window on CR )
 *---------------------------------------------------------------------------*/
int
cr_on_files(void)
{
	return ((int) (beg == end));
}

/*---------------------------------------------------------------------------*
 *	init commandline history system
 *---------------------------------------------------------------------------*/
void
init_history(void)
{
	overstrike = 0;				/* insert is default */

	cur_his = 0;				/* current history pointer */
	strcpy(cbuff, hislines[cur_his]);

	sbeg = cbuff;				/* screen begin at start of buffer */
	beg = cbuff + strlen(cbuff);		/* position beg after prompt */
	pnt = end = beg;			/* point = end = buffer begin */

	kbuf[0] = '\0';				/* init kill - buffer */
	kbuf_init = 1;				/* static kbuf */

	lmax = HISLNLEN - 1;			/* max line length */

	wmove(cmnd_w, C_LINE, 0);		/* beg of line */
	waddstr(cmnd_w, cbuff);			/* write prompt */
}

/*---------------------------------------------------------------------------*
 *	save current commandline to history stack
 *---------------------------------------------------------------------------*/
void
save_line(void)
{
	strncpy(hislines[cur_his], cbuff, HISLNLEN - 1);
	if (cur_his == HISLINES - 1)
		cur_his = 0;
	else
		cur_his++;
	hislines[cur_his][3] = '\0';
	dis_hist();
}

/*---------------------------------------------------------------------------*
 *	get previous history line
 *---------------------------------------------------------------------------*/
void
prev_line(void)
{
	if (cur_his == 0)
		cur_his = HISLINES - 1;
	else
		cur_his--;

	dis_hist();
}

/*---------------------------------------------------------------------------*
 *	get next history line
 *---------------------------------------------------------------------------*/
void
next_line(void)
{
	if (cur_his == HISLINES - 1)
		cur_his = 0;
	else
		cur_his++;

	dis_hist();
}

/*---------------------------------------------------------------------------*
 *	valid new line from history stack, display it
 *---------------------------------------------------------------------------*/
void
dis_hist(void)
{
	strncpy(cbuff, hislines[cur_his], HISLNLEN - 1);

	if ((strlen(cbuff)) < COLS)
	{					/* new line will fit onto screen */
		wmove(cmnd_w, C_LINE, 0);	/* move to start of line */
		wclrtoeol(cmnd_w);		/* clear line */
		wmove(cmnd_w, C_LINE, 0);	/* move to start of line */
		waddstr(cmnd_w, cbuff);		/* write new line */
		sbeg = cbuff;			/* new screen begin */
	}
	else
	{					/* line will not fit onto screen, search end & write new */
		wmove(cmnd_w, C_LINE, 0);	/* start of line */
		wclrtoeol(cmnd_w);		/* clear to end of line */
		sbeg = cbuff;			/* start of buffer */
		while ((strlen(sbeg) + 1) > COLS)
			sbeg += COLS / 2;	/* find start of segment */
		waddstr(cmnd_w, sbeg);		/* write last segment */
	}
	pnt = end = (cbuff + strlen(cbuff));
	beg = cbuff + 3;			/* position beg after prompt */
}

/*---------------------------------------------------------------------------*
 *		move cursor forward character
 *---------------------------------------------------------------------------*/
void
right_line(void)
{
	if (pnt < end)				/* something to move ? */
	{
		pnt++;				/* next char */
		if (pnt > (sbeg + COLS))	/* moving on screen ? */
		{
			sbeg += COLS / 2;	/* new screen begin */
			wmove(cmnd_w, C_LINE, 0);	/* start of screen */
			wclrtoeol(cmnd_w);	/* clear line */
			pmstr(sbeg);		/* write new segment */
		}
	}
}

/*---------------------------------------------------------------------------*
 *		move cursor forward word
 *---------------------------------------------------------------------------*/
void
right_linew(void)
{
	if (pnt < end)				/* something to move ? */
	{
		char *tp;

		if((tp = strpbrk(pnt, " \t")) == NULL)
			return;
			
		pnt = tp+1;			/* first char of next word */
		if (pnt > (sbeg + COLS))	/* moving on screen ? */
		{
			sbeg += COLS / 2;	/* new screen begin */
			wmove(cmnd_w, C_LINE, 0);	/* start of screen */
			wclrtoeol(cmnd_w);	/* clear line */
			pmstr(sbeg);		/* write new segment */
		}
	}
}

/*---------------------------------------------------------------------------*
 *		move cursor back one character
 *---------------------------------------------------------------------------*/
void
left_line(void)
{
	if (pnt > beg)				/* allowed to move ?? */
	{
		pnt--;				/* prev char */
		if (pnt < sbeg)			/* move on screen ?? */
		{
			sbeg -= COLS / 2;	/* back 1/2 length */
			wmove(cmnd_w, C_LINE, 0);	/* start of screen */
			wclrtoeol(cmnd_w);	/* clear line */
			pmstr(sbeg);		/* print this part of buffer */
		}
	}
}

/*---------------------------------------------------------------------------*
 *		move cursor back one word
 *---------------------------------------------------------------------------*/
void
left_linew(void)
{
	if (pnt > beg)				/* allowed to move ?? */
	{
		while((--pnt > beg) && (*pnt == ' ' || *pnt == '\t'))
			;

		while(--pnt > beg)		/* search back */
		{
			if(*pnt == ' ' || *pnt == '\t')
				break;
		}

		if(pnt > beg)
			pnt++;			/* first char of word */

		if (pnt < sbeg)			/* move on screen ?? */
		{
			sbeg -= COLS / 2;	/* back 1/2 length */
			wmove(cmnd_w, C_LINE, 0);	/* start of screen */
			wclrtoeol(cmnd_w);	/* clear line */
			pmstr(sbeg);		/* print this part of buffer */
		}
	}
}

/*---------------------------------------------------------------------------*
 *	append / insert / overstrike a (printable) character
 *---------------------------------------------------------------------------*/
int
cmdline(int c)
{
	if (opt_autoexpand && (c == ' ') && (index(beg, ' ') == NULL))
	{					/* autoexpand=on AND user typed 1st space */
		expand_alias();
	}

	if ((end - beg) < lmax)			/* space in buffer ? */
	{
		if (pnt == end)
/************ append **********/
		{
			*pnt++ = c;		/* store char */
			end++;			/* new end ptr */
			*end = '\0';		/* new terminator */

			if ((end - sbeg + 1) > COLS)
			{
				sbeg += COLS / 2;
				wmove(cmnd_w, C_LINE, 0);
				wclrtoeol(cmnd_w);
				waddstr(cmnd_w, sbeg);
			}
			else
			{
				waddch(cmnd_w, c);	/* echo char */
			}
		}
		else if (!overstrike)
/************** insert **********/
		{
			register char *p;
			register char *q;

			p = end - 1;		/* last char */
			q = end;		/* last char + 1 */
			while (q > pnt)		/* shift right buffer */
				*q-- = *p--;

			*pnt = c;		/* store chr into gap */
			end++;			/* increment end ptr */
			*end = '\0';		/* new terminator */

			if ((pnt + 1) > (sbeg + COLS))	/* scroll right ? */
			{
				sbeg += COLS / 2;	/* new screen beg */
			}
			wmove(cmnd_w, C_LINE, 0);	/* start of line */
			pmstr(sbeg);		/* print string */
			pnt++;			/* new point */
		}
		else
/**************	overstrike ****************/
		{
			*pnt++ = c;		/* store char */

			if ((pnt - sbeg + 1) > COLS)
			{
				sbeg += COLS / 2;
				wmove(cmnd_w, C_LINE, 0);
				wclrtoeol(cmnd_w);
				pmstr(sbeg);
			}
			else
			{
				waddch(cmnd_w, c);	/* echo char */
			}
		}
		return (GOOD);
	}
	else
	{
		flash();			/* no no .. */
		return (BAD);			/* don't accept it */
	}
}

/*---------------------------------------------------------------------------*
 *		delete character left of cursor
 *---------------------------------------------------------------------------*/
void
handlebs(void)
{
	register char *p;
	register char *q;

	if (pnt == beg)				/* at start of buffer ? */
		return;				/* yes */

	q = pnt;				/* next char */
	pnt--;					/* point backward */
	p = pnt;				/* this char */
	while (q < end)				/* shift left buffer */
		*p++ = *q++;
	end--;					/* new end pointer */
	*end = '\0';				/* new end terminator */

	if (pnt < sbeg)				/* moved into previous segment ? */
	{
		sbeg -= COLS / 2;		/* new segment start */
	}
	wmove(cmnd_w, C_LINE, 0);		/* start of screen line */
	wclrtoeol(cmnd_w);			/* clear line */
	pmstr(sbeg);				/* write segment */
}

/*---------------------------------------------------------------------------*
 *		delete character under cursor
 *---------------------------------------------------------------------------*/
void
del_char(void)
{
	register char *p;
	register char *q;

	if (pnt == end)				/* at start of buffer ? */
		return;				/* yes */

	p = pnt;				/* current char */
	q = pnt + 1;				/* next char */

	while (q < end)				/* shift left buffer */
		*p++ = *q++;

	end--;					/* new end pointer */
	*end = '\0';				/* new end terminator */

	wmove(cmnd_w, C_LINE, 0);		/* start of screen line */
	wclrtoeol(cmnd_w);			/* clear line */
	pmstr(sbeg);				/* write new segment */
}

/*---------------------------------------------------------------------------*
 *		print string, maximum length is COLS
 *---------------------------------------------------------------------------*/
static void
pmstr(char *str)
{
	char buf[256];
	register int i = 0;

	while ((*str) && (i < COLS))
		buf[i++] = *str++;
	buf[i] = '\0';
	waddstr(cmnd_w, buf);
}

/*---------------------------------------------------------------------------*
 *		print string, maximum length parameter <n>
 *---------------------------------------------------------------------------*/
static void
pnstr(int n, char *str)
{
	char buf[256];
	register int i = 0;

	while ((*str) && (i < n))
		buf[i++] = *str++;
	buf[i] = '\0';
	waddstr(cmnd_w, buf);
}

/*---------------------------------------------------------------------------*
 * 		delete from cursor to end of line
 *---------------------------------------------------------------------------*/
void
clear_toeol(void)
{
	strcpy(kbuf, pnt);			/* save to kill buffer */
	wclrtoeol(cmnd_w);			/* clear line */
	end = pnt;				/* new end ptr */
	*end = '\0';				/* terminate line */
}

/*---------------------------------------------------------------------------*
 *		move cursor to end of line
 *---------------------------------------------------------------------------*/
void
eol_line(void)
{
	if (pnt == end)				/* already at end ? */
		return;				/* yes */

	if ((strlen(cbuff) < COLS) && (sbeg == cbuff))
	{
		pnt = end;			/* set point to end */
	}
	else
	{
		wmove(cmnd_w, C_LINE, 0);	/* start of line */
		wclrtoeol(cmnd_w);		/* clear to end of line */
		sbeg = cbuff;			/* start of buffer */
		while ((strlen(sbeg) + 1) > COLS)	/* find segment */
			sbeg += COLS / 2;	/*               start */
		waddstr(cmnd_w, sbeg);		/* write last segment */
		pnt = end;			/* new point */
	}
}

/*---------------------------------------------------------------------------*
 * 		move cursor to begin of line
 *---------------------------------------------------------------------------*/
void
bol_line(void)
{
	if (pnt == beg)				/* already at begin of buffer */
		return;				/* nothing to do ! */

	pnt = beg;				/* point = buffer start */

	if (sbeg != cbuff)			/* not in first screen segment */
	{
		wmove(cmnd_w, C_LINE, 0);	/* cursor -> start of screen */
		wclrtoeol(cmnd_w);		/* clear to end of line */
		pmstr(cbuff);			/* write from start of buffer */
		sbeg = cbuff;			/* screen begin = buffer start */
	}
}

/*---------------------------------------------------------------------------*
 *		yank text from kill-buffer to current point
 *---------------------------------------------------------------------------*/
void
yank(void)
{
	if (kbuf[0] == '\0')			/* kill buffer empty ? */
		return;				/* yes, exit */

	if ((end - beg) >= lmax)		/* space in buffer ? */
	{					/* no... */
		if ((overstrike) && (pnt == end))	/* perhaps overstr in the middle ? */
			return;			/* no... */
		if (!overstrike)		/* insert / append ? */
			return;			/* yes */
	}

	/* here if either space in buffer or overwrite not at end */

	if (pnt == end)
/************ append **********/
	{
		/* append kill buffer to buffer, observe buffer max length */

		strncat(cbuff, kbuf, lmax);
		end = cbuff + strlen(cbuff);
		if ((end - sbeg + 1) > COLS)
		{
			while ((end - sbeg + 1) > COLS)
				sbeg += COLS / 2;
			wmove(cmnd_w, C_LINE, 0);
			wclrtoeol(cmnd_w);
			waddstr(cmnd_w, sbeg);	/* write whole new segment */
		}
		else
		{
			waddstr(cmnd_w, pnt);	/* write new end */
		}
		pnt = end;			/* set new pnt */
	}
	else if (!overstrike)
/************** insert **********/
	{
		char tail[HISLNLEN];		/* temp storage */
		register int i;			/* gp... */
		register char *p = pnt;		/* save point pos */

		strcpy(tail, pnt);		/* save string frm point on */
		strncpy(pnt, kbuf, lmax);	/* insert up to max */
		pnt = cbuff + strlen(cbuff);
		strcat(pnt, tail);		/* append saved tail */
		end = cbuff + strlen(cbuff);

		if ((pnt - sbeg + 1) > COLS)
		{				/* new point is outside current segment */
			while ((pnt - sbeg + 1) > COLS)
				sbeg += COLS / 2;
			wmove(cmnd_w, C_LINE, 0);
			wclrtoeol(cmnd_w);
			pmstr(sbeg);		/* print string */
		}
		else
		{				/* new point is inside current segment */
			if ((strlen(sbeg)) < COLS)	/* line > than screen ? */
				i = (sbeg + strlen(sbeg)) - p;	/* no */
			else
				i = (sbeg + COLS) - p;	/* yes */
			pnstr(i, p);		/* new tail */
			p += i;
			while (p-- > pnt)
				wmove(cmnd_w, C_LINE, (curcol() - 1));
		}
	}
	else
/**************	overstrike ****************/
	{
		register int j = lmax;
		register char *k = kbuf;

		while (j-- && *k)
			*pnt++ = *k++;		/* copy kbuf over buffer */

		if (pnt > end)			/* buffer grown bigger ? */
		{
			end = pnt;		/* set new end ptr */
			*end = '\0';		/* terminate buffer */
		}

		if ((pnt - sbeg + 1) > COLS)
		{				/* new point is outside current segment */
			while ((pnt - sbeg + 1) > COLS)
				sbeg += COLS / 2;
			wmove(cmnd_w, C_LINE, 0);
			wclrtoeol(cmnd_w);
			pmstr(sbeg);		/* print string */
		}
		else
		{				/* new point is inside current segment */
			waddstr(cmnd_w, kbuf);
		}
	}
}

/*---------------------------------------------------------------------------*
 *	expand an alias onto the commandline 
 *---------------------------------------------------------------------------*/
void
expand_alias(void)
{
	char buffer[HISLNLEN+16];
	register char *p, *b;

	b = beg;
	p = buffer;

	while(*b && *b != ' ' && *b != '\t')	/* separate first word */
		*p++ = *b++;
	*p = '\0';				/* terminate */

	if ((p = alias(buffer)) != NULL)	/* word = alias ? */
	{
		strcpy(buffer, p);		/* found alias */
		strcat(buffer, b);		/* rest of commandline */
/* XXX check for overflow !!!! */
		strcpy(beg, buffer);		/* buffer -> commandline */
		pnt = end = beg + strlen(beg);	/* (re)set buffer pointers */
		wmove(cmnd_w, C_LINE, 0);	/* update commandline */
		wclrtoeol(cmnd_w);
		waddstr(cmnd_w, sbeg);
	}
}

/*---------------------------------------------------------------------------*
 *	check if the current command is an alias
 *---------------------------------------------------------------------------*/
static char *
alias(char *command)
{
	int i;

	for (i = 0; i < (sizeof(atab) / sizeof(atab[0])) && atab[i][0]; ++i)
		;

	while (i--)
	{
		if (!strcmp(command, atab[i][0]))
			return ((char *) (atab[i][1]));
	}
	return (NULL);
}

/*---------------------------------------------------------------------------*
 *	expand ~ to $HOME
 *---------------------------------------------------------------------------*/
void
expand_tilde(void)
{
	register char *q;
	
	if((q = strchr(beg, '~')) != NULL)
	{
		char tail[HISLNLEN];		/* temp storage */
		register char *p;
		register char *r;
		register int i;			/* gp... */

		pnt = q;			/* pnt is ~ */
		r = envhome;			/* $HOME string */
		
		strcpy(tail, pnt+1);		/* save rest without ~ */
		strncpy(pnt, r, lmax);		/* insert $HOME */
		pnt = cbuff + strlen(cbuff);	/* new point after $HOME */
		strcat(pnt, tail);		/* append saved tail */
		pnt = end = cbuff + strlen(cbuff);	/* new end, new pnt  */

		if ((pnt - sbeg + 1) > COLS)
		{	/* new point is outside current segment */
			while ((pnt - sbeg + 1) > COLS)
				sbeg += COLS / 2;
			wmove(cmnd_w, C_LINE, 0);
			wclrtoeol(cmnd_w);
			pmstr(sbeg);		/* print string */
		}
		else
		{	/* new point is inside current segment */

			p = q;
			
			if ((strlen(sbeg)) < COLS)	/* line > than screen ? */
				i = (sbeg + strlen(sbeg)) - p;	/* no */
			else
				i = (sbeg + COLS) - p;	/* yes */

			wmove(cmnd_w, C_LINE, p-sbeg);
			pnstr(i, p);		/* new tail */
		}
	}
}

/*----------------------- E O F -------------------------------------------*/
