/*
 *
 * Copyright (c) 1988, 1999 Hellmuth Michaelis. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *---------------------------------------------------------------------------
 *
 *	paging.c - file window page up/down
 *      -----------------------------------
 *
 *	Last Edit-Date: [Fri Jul 30 11:04:14 1999]
 *
 *	$Id$
 *
 *----------------------------------------------------------------------------*/

#include "cwish.h"

static char curpath[MAXPATHLEN + 256];		/* current path buffer normalized for screen */
static int lastpage;				/* number of last page */

static void make_cpath();			/* get & format current directory */
static void dis_cpath(int init);		/* print current directory & entries */
static void new_stops();
static void dis_init(struct onefile *);
static void dis_play(struct onefile *);
static void show_blip(void);

/*---------------------------------------------------------------------------*
 *	get & format current directory for display
 *---------------------------------------------------------------------------*/
static void
make_cpath(void)
{
	int i, j, width;
	static int firsttime = 1;

	if ((getcwd(cur_path, MAXPATHLEN)) == NULL)
	{
		error("internal getwcd() error, chdir to HOME");
		if ((cd(envhome)) == -1)
			fatal("cannot chdir() to HOME");
		if ((getcwd(cur_path, MAXPATHLEN)) == NULL)
			fatal("twotimes internal getcwd() error");
	}

	if (firsttime)
	{
		strcpy(lcur_path, cur_path);	/* symlink path */
		firsttime = 0;
	}

	if (opt_followsdir)
		strcpy(curpath, lcur_path);	/* display symlink path */
	else
		strcpy(curpath, cur_path);	/* display hard path */

	width = COLS - CURDIRW - (ENTRYSW + ENTRYCW) - 1;
	i = strlen(curpath);
	if (i > width)
	{
		j = 0;
		i = i - width + 4;
		while (curpath[i++] != '/')
			;
		curpath[j++] = '.';
		curpath[j++] = '.';
		curpath[j++] = '.';
		curpath[j++] = ' ';
		i--;
		while (curpath[i])
			curpath[j++] = curpath[i++];
		curpath[j++] = '\0';
	}
}

/*---------------------------------------------------------------------------*
 *	print current directory & entries
 *---------------------------------------------------------------------------*/
static void
dis_cpath(int init)
{
	static int p_init = 1;
	int i;

	if (p_init || init)
	{
		wmove(fst_w, 0, 0);		/* move to dir string */
		wattron(fst_w, A_UNDERLINE);	/* underline on */
		waddstr(fst_w, "Directory:");	/* description */
		wattroff(fst_w, A_UNDERLINE);	/* end underline */
	}
	wmove(fst_w, 0, CURDIRW);
	waddstr(fst_w, curpath);		/* print current path */

	i = CURDIRW + strlen(curpath);

	while (i++ < COLS - ENTRYSW - ENTRYCW)
		waddch(fst_w, SPACE);

	if (p_init || init)
	{
		/* move to entries string */
		wmove(fst_w, 0, COLS - ENTRYSW - ENTRYCW);
		wclrtoeol(fst_w);
		wattron(fst_w, A_UNDERLINE);	/* underline on */

		if (tagsize)	/* tagged files, display sum of sizes */
		{
			waddstr(fst_w, "Byte:");	/* description */
			wmove(fst_w, 0, COLS - ENTRYCW - 4); /* move to entries string */
			wattroff(fst_w, A_UNDERLINE);	/* underline off */
			wprintw(fst_w, "%8d", tagsize);	/* print size */
		}
		else if (opt_wildon)		/* wildcarding enabled */
		{
			waddstr(fst_w, "Matches:");	/* description */
			wmove(fst_w, 0, COLS - ENTRYCW);/* move to entries string */
			wattroff(fst_w, A_UNDERLINE);	/* underline off */
			waddstr(fst_w, counter);	/* print current path */
		}
		else				/* normal operation */
		{
			waddstr(fst_w, "Entries:");	/* description */
			wmove(fst_w, 0, COLS - ENTRYCW);/* move to entries string */
			wattroff(fst_w, A_UNDERLINE);	/* underline off */
			waddstr(fst_w, counter);	/* print current path */
		}
		p_init = 0;
	}
}

/*---------------------------------------------------------------------------*
 *	calculate new "tab" stops for filename display
 *---------------------------------------------------------------------------*/
static void
new_stops(void)
{
	int j = STARTCOL;			/* start value */
	int maxcol = 0;				/* global no of cols */

	while (maxcol < STOPS)
	{
		stops[maxcol++] = j;		/* set stop point */
		j += MINSPACE;			/* add minimum space count between names */
		j += maxfnleng;			/* and next filename's length */
		if ((j + MINSPACE + maxfnleng) > COLS)
		{
			stops[maxcol] = 0;	/* end marker */
			break;			/* leave loop */
		}
	}
}

/*---------------------------------------------------------------------------*
 *	display next file page (if any) in file window
 *---------------------------------------------------------------------------*/
void
next_page(void)
{
	if (cur_page >= lastpage)		/* already on last page ? */
		return;				/* yes, leave .. */

	wclear(file_w);				/* clear current window */
	dis_cpath(0);				/* current path & entries */

	for (;;)
	{
		if ((cur_file->page & PAGE_MSK) == ((cur_page + 1) & PAGE_MSK))
			break;
		if (cur_file->next != NULL)
			cur_file = cur_file->next;	/* next entry */
		else
			break;
	}

	cur_page = cur_file->page & PAGE_MSK;

	scr_beg = cur_file;
	scr_end = cur_file;

	dis_play(cur_file);			/* display directories */
	show_blip();
	cur_inv(cur_file);
}

/*---------------------------------------------------------------------------*
 *	display previous file page (if any) in file window
 *---------------------------------------------------------------------------*/
void
prev_page(void)
{
	if (cur_page <= 0)			/* is there a previous page ? */
		return;				/* no, return */

	wclear(file_w);				/* clear current window */
	dis_cpath(0);				/* display current path & entries */

	cur_file = first;

	for (;;)
	{
		if ((cur_file->page & PAGE_MSK) == ((cur_page - 1) & PAGE_MSK))
			break;
		if (cur_file->next != NULL)
			cur_file = cur_file->next;	/* next entry */
		else
			break;
	}

	cur_page = cur_file->page & PAGE_MSK;

	scr_beg = cur_file;
	scr_end = cur_file;

	dis_play(cur_file);			/* display directories */
	show_blip();
	cur_inv(cur_file);
}

/*---------------------------------------------------------------------------*
 *	re-fill/refresh file window with data for current dir
 *---------------------------------------------------------------------------*/
void
fresh_files(void)
{
	scr_end = scr_beg;			/* init */
	wclear(file_w);				/* clear current window */
	dis_cpath(1);				/* display current path */
	dis_play(scr_beg);			/* display directories */
	show_blip();				/* show more indicator */
	cur_inv(cur_file);			/* highlight old */
}

/*---------------------------------------------------------------------------*
 *	initialize the file display window
 *---------------------------------------------------------------------------*/
void
init_files(int preserve, char *dirname)
{
	char filename[MAXPATHLEN];
	int page;
	int row;
	int col;

	if (cur_file != NULL)
	{
		cur_blink(cur_file);		/* blink current file while reading */
		wrefresh(file_w);		/* update blinking */
		page = cur_file->page & PAGE_MSK;
		row = cur_file->orow;
		col = cur_file->ocol;
	}
	else
	{
		page = row = col = 0;
	}

	if (preserve == PRES_NORM)
	{
		strcpy(filename, cur_file->onam);
		page = cur_file->page & PAGE_MSK;
		row = cur_file->orow;
		col = cur_file->ocol;
	}
	else if (preserve == PRES_DD)
	{
		if (dirname != NULL && *dirname && opt_preserve)
		{
			char *ptr;

			if ((ptr = (char *) rindex(dirname, '/')) == NULL)
			{
				preserve = PRES_NO;
			}
			else if (*dirname)
			{
				ptr++;
				strcpy(filename, ptr);
				page = 0;
				row = 0;
				col = 0;
			}
			else
			{
				preserve = PRES_NO;
			}
		}
		else
		{
			preserve = PRES_NO;
		}
	}

	cur_page = 0;				/* file display page 0 */
	make_cpath();				/* format current dir */

	if (fill_list() != 0)			/* get directory data into list */
	{
		if (first != NULL)		/* if any malloced space ... */
			free_list();		/* FREE IT !!! */
		return;				/* return with error */
	}

	dis_cpath(1);				/* display current path */
	new_stops();				/* compute new stops-table */
	cur_file = first;			/* current file is first */
	scr_beg = first;			/* first filename on display */
	scr_end = first;			/* last filename on display */
	wclear(file_w);				/* clear current window */

	while (cur_file != NULL)		/* init whole list */
	{
		dis_init(cur_file);		/* page init directories */
		cur_page++;			/* next page */
		cur_file = scr_end->next;	/* new pointer */
		scr_beg = cur_file;		/* new screen start */
	}
	lastpage = cur_page - 1;		/* init last page no. */
	cur_page = 0;				/* file display page 0 */
	cur_file = first;			/* current file is first */
	scr_beg = first;			/* first filename on display */
	scr_end = first;			/* last filename on display */

	if (preserve == PRES_NORM || preserve == PRES_DD)
	{
		int found = 0;
		int length = strlen(filename);
		struct onefile *last_file = cur_file;

		while (!found)
		{
			if (preserve == PRES_NORM)
			{
				if (!(strncmp(filename, last_file->onam, length)))
				{
					found = 1;
					break;
				}
			}
			else if (preserve == PRES_DD)
			{
				if (!(strncmp(filename, &(last_file->onam[1]), length)))
				{
					found = 1;
					break;
				}
			}

			if (last_file->next)
				last_file = last_file->next;
			else
				break;
		}
		if (found)
		{
			page = ((last_file->page) & PAGE_MSK);
			row = last_file->orow;
			col = last_file->ocol;
		}
		if (page > lastpage)
			page = lastpage;

		for (;;)
		{
			if ((cur_file->page & PAGE_MSK) == (page & PAGE_MSK))
				break;
			if (cur_file->next != NULL)
				cur_file = cur_file->next;
			else
				break;
		}
		cur_page = cur_file->page & PAGE_MSK;

		scr_beg = cur_file;
		scr_end = cur_file;

		if (found)
			cur_file = last_file;
		else
		{
			last_file = cur_file;

			while ((last_file->page & PAGE_MSK) == (page & PAGE_MSK))
			{
				if ((last_file->orow == row) && (last_file->ocol == col))
				{
					cur_file = last_file;
					break;
				}
				if (last_file->next != NULL)
					last_file = last_file->next;
				else
					break;
			}
		}
	}
	dis_play(scr_beg);			/* display directories */
	show_blip();
	cur_inv(cur_file);			/* handle current filename */
}

/*---------------------------------------------------------------------------*
 *	display file window
 *---------------------------------------------------------------------------*/
static void
dis_play(register struct onefile *n_ptr)
{
	int displaytype = 1;			/* display file type */
	int what = n_ptr->onam[0];		/* dirs, files, executables */
	int curp = n_ptr->page & PAGE_MSK;

	while (n_ptr)
	{
		if (displaytype)		/* print description */
		{
			wmove(file_w, n_ptr->orow, 0);	/* move type pos. */
			wattron(file_w, A_UNDERLINE);	/* underline */
			switch (what)
			{
				case ISDIR:	/* directories */
					waddstr(file_w, "Dirs:");
					break;

				case ISPROG:	/* executables */
					waddstr(file_w, "Progs:");
					break;

				case ISDATA:	/* data */
					waddstr(file_w, "Files:");
					break;

				case ISCHAR:	/* char devs */
					waddstr(file_w, "Cdevs:");
					break;

				case ISBLOCK:	/* block devs */
					waddstr(file_w, "Bdevs:");
					break;

				case ISPIPE:	/* pipes */
					waddstr(file_w, "Pipes:");
					break;

				case ISSOCK:	/* sockets */
					waddstr(file_w, "Sockt:");
					break;

				default:	/* data */
					waddstr(file_w, "Unkwn:");
					break;
			}
			wattroff(file_w, A_UNDERLINE);	/* end underline */
			displaytype = 0;	/* reset displaytype */
		}

		is_tagged(n_ptr);		/* print it */

		scr_end = n_ptr;		/* new last screen pos */

		if (!n_ptr->next)		/* next entry valid ?? */
			break;			/* no, current is last */

		n_ptr = n_ptr->next;		/* next entry */

		if (n_ptr->page != curp)	/* if next page != current page .. */
			break;			/* ..exit */

		if (n_ptr && (what != n_ptr->onam[0]))	/* new type ?? */
		{
			what = n_ptr->onam[0];	/* new type */
			displaytype = 1;	/* display */
		}
	}
}

/*---------------------------------------------------------------------------*
 *	initialize file window display
 *---------------------------------------------------------------------------*/
static void
dis_init(register struct onefile *n_ptr)
{
	int firstpage = 1;			/* mark first page flag */
	int row = 0;				/* current row */
	int what = n_ptr->onam[0];		/* dirs, files, executables */
	int stop = 0;				/* first stop */
	int col = stops[stop];			/* current column */

	while (n_ptr && (row < fileheight))	/* while not at end of list */
		/*  and in file window .... */
	{
		if (firstpage)			/* mark first file first page */
		{
			n_ptr->page = (cur_page | FIRST);
			firstpage = 0;
		}
		else
		{
			n_ptr->page = cur_page;
		}

		n_ptr->orow = row;		/* save current pos */
		n_ptr->ocol = col;		/*  into list entry */

		stop++;				/* next stop */
		if (stops[stop] == 0)		/* last stop ? */
		{
			row++;			/* next row */
			stop = 0;		/* new stop */
		}
		col = stops[stop];		/* new col */

		scr_end = n_ptr;		/* new last screen pos */

		n_ptr = n_ptr->next;		/* next entry */

		if (n_ptr && (what != n_ptr->onam[0]))	/* new type ?? */
		{
			what = n_ptr->onam[0];	/* new type */

			if (stop == 0)		/* compute vertical space */
			{
				row++;
			}
			else
			{
				row += 2;
				stop = 0;
			}
			col = stops[stop];	/* new column */
		}
	}
}

/*---------------------------------------------------------------------------*
 *	show wether we have a previous(^), next(v) or just one(*) page
 *---------------------------------------------------------------------------*/
static void
show_blip(void)
{
	if ((!cur_page) && (!lastpage))		/* just one page */
		return;

	wmove(file_w, fileheight - 1, COLS - 1);

	wattrset(file_w, A_REVERSE);

	if ((!cur_page) && (lastpage))		/* first page */
	{
		waddch(file_w, 'v');
	}
	else if (lastpage && (cur_page == lastpage))	/* last page */
	{
		waddch(file_w, '^');
	}
	else
		/* somewhere between */
	{
		waddch(file_w, '*');
	}
	wattrset(file_w, A_NORMAL);
}

/*---------------------------------- EOF -------------------------------------*/
